"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides a direct access and management logic of the administration file
(:code:`info.json´).

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020 Dead Hosts
    Copyright (c) 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import logging
from datetime import datetime, timedelta

import PyFunceble.helpers as pyfunceble_helpers

from .configuration import Paths
from .travis_ci import TravisCI


class Info:
    """
    Provides an interface to manage the `info.json` file.

    :ivar str working_directory:
        Our working directory.
    :ivar str file:
        The full path of the previously known as `info.json` file.
    :ivar file_instance:
        An instance of :py:class`PyFunceble.helpers.File`.
    :vartype file_instance: :py:class`PyFunceble.helpers.File
    :ivar dict content:
        The content of `info.json` (in dict).
    """

    def __init__(self):
        self.working_directory: str = TravisCI.get_working_dir()

        self.file: str = self.working_directory + Paths.info_filename
        self.file_instance: pyfunceble_helpers.File = pyfunceble_helpers.File(self.file)

        if self.file_instance.exists():
            self.content: dict = pyfunceble_helpers.Dict().from_json(
                self.file_instance.read()
            )
        else:
            self.content: dict = {}

        logging.debug("Administration file path: %s", self.file)
        logging.debug("Administration file exists: %s", self.file_instance.exists())
        logging.debug("Administration file content: %s", self.file_instance.read())
        logging.debug("Administration file interpreted: %s", self.content)

        self.filter()
        self.create_missing_index()
        self.clean()
        self.save()

    def __getitem__(self, index):
        if index in self.content:
            return self.content[index]
        raise KeyError(index)

    def __getattr__(self, index):
        if index in self.content:
            return self.content[index]
        raise AttributeError(index)

    def __setitem__(self, index, value):
        self.content[index] = value
        self.save()

    def filter(self):
        """
        Processes a first filtering of the file.
        """

        # pylint: disable=too-many-branches

        self.content["name"] = Paths.git_base_name

        logging.debug("Updated the `name` index of the administration file.")

        to_delete = [
            pyfunceble_helpers.File(self.working_directory + ".administrators"),
            pyfunceble_helpers.File(self.working_directory + "update_me.py"),
            pyfunceble_helpers.File(self.working_directory + "admin.py"),
        ]

        if "list_name" in self.content:
            to_delete.append(
                pyfunceble_helpers.File(
                    self.working_directory + self.content["list_name"]
                )
            )

        if "ping" in self.content:
            local_ping_result = []
            for username in self.content["ping"]:
                if username.startswith("@"):
                    local_ping_result.append(username)
                else:
                    local_ping_result.append(f"@{username}")

            self.content["ping"] = local_ping_result
            logging.debug(
                "Updated the `ping` index of the administration file, "
                "the format has to stay the same everywhere."
            )

        if (
            "raw_link" in self.content
            and isinstance(self.content["raw_link"], str)
            and not self.content["raw_link"]
        ):
            self.content["raw_link"] = None
            logging.debug(
                "Updated the `raw_link` index of the administration file, "
                "empty string not accepted."
            )

        if (
            "custom_pyfunceble_config" in self.content
            and self.content["custom_pyfunceble_config"]
            and not isinstance(self.content["custom_pyfunceble_config"], dict)
        ):
            self.content["custom_pyfunceble_config"] = {}
            logging.debug(
                "Updated the `custom_pyfunceble_config` index of the "
                "administration file, it should be a %s.",
                dict,
            )

        for file_to_delete in to_delete:
            if file_to_delete.exists():
                file_to_delete.delete()

                logging.debug(
                    "Deleted the %s file, it is not needed anymore.",
                    repr(file_to_delete.path),
                )

        for index in ["currently_under_test"]:
            if index in self.content and not isinstance(self.content[index], bool):
                self.content[index] = bool(int(self.content[index]))

                logging.debug(
                    "Updated the %s index of the administration file, "
                    "it should be a %s.",
                    repr(index),
                    bool,
                )

        for index in [
            "days_until_next_test",
            "finish_timestamp",
            "last_download_datetime"
            "last_download_timestamp"
            "lastest_part_finish_timestamp",
            "lastest_part_start_timestamp",
            "start_timestamp",
        ]:
            if index in self.content and not isinstance(self.content[index], float):
                self.content[index] = float(self.content[index])
                logging.debug(
                    "Updated the %s index of the administration file, "
                    "it should be a %s.",
                    repr(index),
                    float,
                )

        for index in [
            "finish_timestamp",
            "last_download_timestamp",
            "lastest_part_finish_timestamp",
            "lastest_part_start_timestamp",
            "start_timestamp",
        ]:
            if index in self.content and not isinstance(self.content[index], datetime):
                self.content[index] = datetime.fromtimestamp(self.content[index])
                logging.debug(
                    "Updated the %s index of the administration file, "
                    "the system understands %s only."
                    " (JSON => %s).",
                    repr(index),
                    datetime,
                    dict,
                )

        for index in [
            "finish_datetime",
            "last_download_datetime",
            "lastest_part_finish_datetime",
            "lastest_part_start_datetime",
            "start_datetime",
        ]:
            if index in self.content:
                if self.content[index] and not isinstance(
                    self.content[index], datetime
                ):
                    self.content[index] = datetime.fromisoformat(self.content[index])
                    logging.debug(
                        "Updated the %s index of the administration file, "
                        "the system understands %s only."
                        " (JSON => %s.",
                        repr(index),
                        datetime,
                        dict,
                    )
                else:
                    self.content[index] = datetime.fromtimestamp(0)
                    logging.debug(
                        "Set the %s index of the administration file, "
                        "it was not previously set.",
                        repr(index),
                    )

    def create_missing_index(self):
        """
        Creates the missing indexes.
        """

        default_datetime = datetime.now() - timedelta(days=10)

        indexes = {
            "currently_under_test": False,
            "custom_pyfunceble_config": {},
            "days_until_next_test": 2,
            "finish_datetime": default_datetime,
            "finish_timestamp": default_datetime.timestamp(),
            "last_download_datetime": default_datetime,
            "last_download_timestamp": default_datetime.timestamp(),
            "latest_part_finish_timestamp": default_datetime.timestamp(),
            "latest_part_start_timestamp": default_datetime.timestamp(),
            "latest_part_finish_datetime": default_datetime,
            "latest_part_start_datetime": default_datetime,
            "name": Paths.git_base_name,
            "own_management": False,
            "ping": [],
            "raw_link": None,
            "start_datetime": default_datetime,
            "start_timestamp": default_datetime.timestamp(),
            "live_update": True,
        }

        for index, value in indexes.items():
            if index not in self.content:
                self.content[index] = value
                logging.debug(
                    "Created the %s index of the administration file, it was not found.",
                    repr(index),
                )

    def clean(self):
        """
        Removes the unneeded indexes.
        """

        for index in [
            "arguments",
            "clean_list_file",
            "clean_original",
            "commit_autosave_message",
            "last_test",
            "list_name",
            "stable",
        ]:
            if index in self.content:
                del self.content[index]

                logging.debug(
                    "Deleted the %s index of the administration file, "
                    "it is not needed anymore.",
                    repr(index),
                )

    def save(self):
        """
        Saves the current version of the content file.
        """

        origin = self.content.copy()

        for index, value in self.content.copy().items():
            if index.endswith("_timestamp") and isinstance(value, datetime):
                self.content[index] = value.timestamp()
            elif index.endswith("_datetime") and isinstance(value, datetime):
                self.content[index] = value.isoformat()

        pyfunceble_helpers.Dict(self.content).to_json_file(self.file)
        self.content = origin

    def get_ping_for_commit(self):
        """
        Returns the list of username for the commit message.
        """

        if "ping" in self.content:
            return " ".join(self.content["ping"])
        return ""



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the orchestration logic. In other words, put everything
together to process a test.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020 Dead Hosts
    Copyright (c) 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import logging
import sys
from datetime import datetime
from os import environ

import PyFunceble
import PyFunceble.helpers as pyfunceble_helpers
from colorama import Fore, Style
from colorama import init as initiate_colorama

from .. import updater
from ..configuration import Paths
from ..configuration import TravisCI as TravisCIConfig
from ..helpers import Command
from ..info import Info
from ..travis_ci import TravisCI
from .authorize import Authorize


class Orchestration:
    """
    Puts everything together.
    """

    def __init__(self, save=False, end=False):
        initiate_colorama(autoreset=True)

        pyfunceble_helpers.EnvironmentVariable(
            "PYFUNCEBLE_AUTO_CONFIGURATION"
        ).set_value("YES")

        git_name = pyfunceble_helpers.EnvironmentVariable("GIT_NAME")
        git_email = pyfunceble_helpers.EnvironmentVariable("GIT_EMAIL")

        if git_email.exists() and "funilrys" in git_email.get_value():
            git_name.set_value("Dead-Hosts")
            git_email.set_value(TravisCIConfig.default_email)

        self.info_manager = Info()
        self.working_directory = self.info_manager.working_directory

        logging.info("Working directory: %s", self.working_directory)

        self.authorize = Authorize(self.info_manager)

        self.origin_file_instance = PyFunceble.helpers.File(
            self.working_directory + Paths.origin_filename
        )
        self.input_file_instance = PyFunceble.helpers.File(
            self.working_directory + Paths.input_filename
        )
        self.output_file_instance = PyFunceble.helpers.File(
            self.working_directory + Paths.output_filename
        )

        logging.info("Origin file: %s", self.origin_file_instance.path)
        logging.info("Input file: %s", self.input_file_instance.path)
        logging.info("Output file: %s", self.output_file_instance.path)

        if not end and not save:
            logging.info("Checking authorization to run.")

            if self.authorize.test():
                updater.exec_all()

                PyFunceble.load_config(generate_directory_structure=False)
                self.fetch_list_to_test()

                self.run_test()
            else:
                logging.info(
                    "Not authorized to run a test until %s " "(current time) > %s",
                    datetime.now(),
                    self.authorize.get_test_authorization_time(),
                )
                sys.exit(0)
        elif save:
            self.run_autosave()
        else:
            self.run_end()

    def fetch_list_to_test(self):
        """
        Provides the latest version of the file to test.
        """

        result = {}
        if self.authorize.refresh():
            logging.info("We are authorized to refresh the lists! Let's do that.")
            logging.info("Raw Link: %s", self.info_manager.raw_link)

            if self.info_manager.raw_link:
                result["origin"] = PyFunceble.helpers.Download(
                    self.info_manager.raw_link
                ).text()

                logging.info(
                    "Could get the new version of the list. Updating the download time."
                )

                self.info_manager["last_download_datetime"] = datetime.now()
                self.info_manager["last_download_timestamp"] = self.info_manager[
                    "last_download_datetime"
                ].timestamp()
            elif self.origin_file_instance.exists():
                logging.info(
                    "Raw link not given or is empty. Let's work with %s.",
                    self.origin_file_instance.path,
                )

                result["origin"] = self.origin_file_instance.read()

                logging.info("Emptying the download time.")

                self.info_manager["last_download_datetime"] = datetime.fromtimestamp(0)
                self.info_manager["last_download_timestamp"] = self.info_manager[
                    "last_download_datetime"
                ].timestamp()
            else:
                logging.info(
                    f"Could not find {self.origin_file_instance.path}. "
                    "Generating empty content to test."
                )

                result["origin"] = ""

                logging.info("Emptying the download time.")

                self.info_manager["last_download_datetime"] = datetime.fromtimestamp(0)
                self.info_manager["last_download_timestamp"] = self.info_manager[
                    "last_download_datetime"
                ].timestamp()

            if not self.info_manager["custom_pyfunceble_config"]:
                logging.info("Formatting the list to test.")

                result["input_list"] = []

                for line in result["origin"].splitlines():
                    converted = PyFunceble.converter.File(line).get_converted()

                    if not converted:
                        continue

                    if isinstance(converted, list):
                        result["input_list"].extend([x for x in converted if x])
                    else:
                        result["input_list"].append(converted)
            else:
                result["input_list"] = result["origin"].splitlines()

            result["input"] = "\n".join(result["input_list"])

            self.origin_file_instance.write(result["origin"], overwrite=True)
            self.input_file_instance.write(result["input"], overwrite=True)

            logging.info("Updated %s.", self.origin_file_instance.path)
            logging.info("Updated %s.", self.input_file_instance.path)

        return result

    def run_test(self):
        """
        Run a test of the input list.
        """

        if not self.info_manager.currently_under_test:
            self.info_manager["currently_under_test"] = True

            self.info_manager["start_datetime"] = datetime.now()
            self.info_manager["start_timestamp"] = self.info_manager[
                "start_datetime"
            ].timestamp()

            self.info_manager["finish_datetime"] = datetime.fromtimestamp(0)
            self.info_manager["finish_timestamp"] = self.info_manager[
                "finish_datetime"
            ].timestamp()

        self.info_manager["latest_part_start_datetime"] = datetime.now()
        self.info_manager["latest_part_start_timestamp"] = self.info_manager[
            "latest_part_start_datetime"
        ].timestamp()

        self.info_manager["latest_part_finish_datetime"] = datetime.fromtimestamp(0)
        self.info_manager["latest_part_finish_timestamp"] = self.info_manager[
            "latest_part_finish_datetime"
        ].timestamp()

        logging.info("Updated all timestamps.")
        logging.info("Starting PyFunceble %s ...", PyFunceble.VERSION)

        Command(f"PyFunceble -f {self.input_file_instance.path}").run_to_stdout()

        if not TravisCIConfig.github_token or not TravisCIConfig.build_dir:
            self.run_end()

    def run_autosave(self):
        """
        Run the autosave logic of the administration file.

        .. warning::
            This is just about the administration file not PyFunceble.
        """

        self.info_manager["latest_part_finish_datetime"] = datetime.now()
        self.info_manager["latest_part_finish_timestamp"] = self.info_manager[
            "latest_part_finish_datetime"
        ].timestamp()

        logging.info("Updated all timestamps.")

        PyFunceble.helpers.Directory(self.working_directory + "db_types").delete()

    def run_end(self):
        """
        Run the end logic.
        """

        self.info_manager["currently_under_test"] = False

        self.info_manager["latest_part_finish_datetime"] = datetime.now()
        self.info_manager["latest_part_finish_timestamp"] = self.info_manager[
            "latest_part_finish_datetime"
        ].timestamp()

        self.info_manager["finish_datetime"] = self.info_manager[
            "latest_part_finish_datetime"
        ]
        self.info_manager["finish_timestamp"] = self.info_manager[
            "finish_datetime"
        ].timestamp()

        logging.info("Updated all timestamps and indexes that needed to be updated.")

        pyfunceble_active_list = PyFunceble.helpers.File(
            self.working_directory + "output/domains/ACTIVE/list"
        )
        clean_list = [
            "# File generated by the Dead-Hosts project with the help of PyFunceble.",
            "# Dead-Hosts: https://github.com/dead-hosts",
            "# PyFunceble: https://pyfunceble.github.io",
            f"# Generation Time: {datetime.now().isoformat()}",
        ]

        logging.info(f"PyFunceble ACTIVE list output: {pyfunceble_active_list.path}")

        if pyfunceble_active_list.exists():
            logging.info(
                f"{pyfunceble_active_list.path} exists, getting and formatting its content."
            )

            clean_list.extend(
                PyFunceble.helpers.List(
                    PyFunceble.helpers.Regex(r"^#").get_not_matching_list(
                        pyfunceble_active_list.read().splitlines()
                    )
                ).format()
            )

        self.output_file_instance.write("\n".join(clean_list), overwrite=True)
        logging.info("Updated of the content of %s", self.output_file_instance.path)

        PyFunceble.helpers.Directory(self.working_directory + "db_types").delete()



"""
Dead Hosts's launcher - The launcher of the Dead-Hosts infrastructure.

Provides the updater of the README file.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Project link:
    https://github.com/dead-hosts/infrastructure-launcher

License:
::

    MIT License

    Copyright (c) 2019, 2020 Dead Hosts
    Copyright (c) 2019, 2020 Nissar Chababy

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import logging

import PyFunceble.helpers as pyfunceble_helpers

from ..configuration import Markers, Paths
from .base import Base


class ReadMeUpdater(Base):
    """
    Provides the updater of the README file.
    """

    def __init__(self) -> None:
        self.destination = pyfunceble_helpers.File(Paths.readme_filename)

        super().__init__()

    def authorization(self) -> bool:
        return self.destination.exists()

    def pre(self):
        logging.info("Started to update the content of %s!", self.destination.path)

    def post(self):
        logging.info("Finished to update the content of %s!", self.destination.path)

    def start(self) -> None:
        logging.info(
            "Started to update the `About PyFunceble` section of %s",
            self.destination.path,
        )

        updated_version = pyfunceble_helpers.Regex(
            Markers.extract_about_pyfunceble
        ).replace_match(self.destination.read(), Markers.about_pyfunceble)

        logging.info(
            "Finished to update the `About PyFunceble` section of %s",
            self.destination.path,
        )

        logging.info(
            "Started to update the `About Dead-Hosts` section of %s",
            self.destination.path,
        )

        updated_version = pyfunceble_helpers.Regex(
            Markers.extract_about_dead_hosts
        ).replace_match(self.destination.read(), Markers.about_dead_hosts)

        logging.info(
            "Finished to update the `About Dead-Hosts` section of %s",
            self.destination.path,
        )

        self.destination.write(updated_version, overwrite=True)
